/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * \package    RobotAPI::gui-plugins::SkillManagerMonitorWidgetController
 * \author     Raphael Grimm ( raphael dot grimm at kit dot edu )
 * \date       2020
 * \copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "SkillManagerMonitorWidgetController.h"

#include <memory>
#include <optional>
#include <regex>
#include <string>

#include <SimoxUtility/algorithm/string.h>

#include <RobotAPI/libraries/skills/core/Skill.h>

#include "aronTreeWidget/visitors/AronTreeWidgetConverter.h"
#include "aronTreeWidget/visitors/AronTreeWidgetModalCreator.h"

// modals
#include "aronTreeWidget/modal/text/AronTreeWidgetTextInputModalController.h"

// debug
#include <QAction>
#include <QClipboard>
#include <QDoubleSpinBox>
#include <QGridLayout>
#include <QMenu>
#include <QTextBrowser>

#include <RobotAPI/libraries/aron/converter/json/NLohmannJSONConverter.h>
#include <RobotAPI/libraries/skills/core/Skill.h>
#include <RobotAPI/libraries/skills/core/SkillExecutionRequest.h>

#include "aronTreeWidget/Data.h"
#include "aronTreeWidget/modal/text/AronTreeWidgetTextInputModalController.h"
#include "aronTreeWidget/visitors/AronTreeWidgetCreator.h"
#include "aronTreeWidget/visitors/AronTreeWidgetModalCreator.h"
#include "aronTreeWidget/widgets/SkillDescriptionWidget.h"

//configSk
namespace armarx
{
    const skills::SkillID SkillManagerMonitorWidgetController::SelectedSkill::UNK_SKILL_ID =
        skills::SkillID{.providerId = ::std::nullopt, .skillName = skills::SkillID::UNKNOWN};

    QPointer<QDialog>
    SkillManagerMonitorWidgetController::getConfigDialog(QWidget* parent)
    {
        if (!dialog)
        {
            dialog = new SimpleConfigDialog(parent);
            dialog->addProxyFinder<skills::manager::dti::SkillManagerInterfacePrx>(
                "SkillMemory", "", "SkillMem*");
        }
        return qobject_cast<SimpleConfigDialog*>(dialog);
    }

    void
    SkillManagerMonitorWidgetController::configured()
    {
        observerName = dialog->getProxyName("SkillMemory");
    }

    void
    SkillManagerMonitorWidgetController::loadSettings(QSettings* settings)
    {
        observerName = settings->value("SkillMemory", "SkillMemory").toString().toStdString();
    }

    void
    SkillManagerMonitorWidgetController::saveSettings(QSettings* settings)
    {
        settings->setValue("SkillMemory", QString::fromStdString(observerName));
    }
} // namespace armarx

// Others
namespace armarx
{
    /*
    void
    SkillManagerMonitorWidgetController::prepareAndRunMenu(const QPoint& pos)
    {
        QMenu* menu = new QMenu();

        // Stop skill
        QAction* stopSkillAction = new QAction("Stop execution", this);
        skills::SkillStatus currentStatus =
            skillStatusUpdates.at(selectedSkill.skillExecutionId).status;
        stopSkillAction->setDisabled(currentStatus == skills::SkillStatus::Aborted ||
                                     currentStatus == skills::SkillStatus::Failed ||
                                     currentStatus == skills::SkillStatus::Succeeded);

        QAction* rerunSkillAction = new QAction("Re-run with similar params", this);
        menu->addAction(stopSkillAction);
        menu->addAction(rerunSkillAction);
        connect(stopSkillAction,
                &QAction::triggered,
                this,
                &SkillManagerMonitorWidgetController::stopSkill);
        connect(rerunSkillAction,
                &QAction::triggered,
                this,
                &SkillManagerMonitorWidgetController::rerunSkillWithSimilarParams);

        // Temporarily disable rerun-skill-Action
        rerunSkillAction->setDisabled(true);

        // open menu
        menu->popup(widget.treeWidgetSkillExecutions->viewport()->mapToGlobal(pos));
    }

    SkillExecutionInfoTreeWidgetItem*
    SkillExecutionInfoTreeWidgetItem::SearchRecursiveForMatch(
        SkillExecutionInfoTreeWidgetItem* haystack,
        const skills::SkillExecutionID& needle)
    {
        if (!haystack)
        {
            return nullptr;
        }

        if (needle == haystack->executionId)
        {
            return haystack;
        }
        for (int i = 0; i < haystack->childCount(); ++i)
        {
            auto el = static_cast<SkillExecutionInfoTreeWidgetItem*>(haystack->child(i));
            return SkillExecutionInfoTreeWidgetItem::SearchRecursiveForMatch(el, needle);
        }
        return nullptr;
    }
*/

    SkillManagerMonitorWidgetController::SkillManagerMonitorWidgetController()
    {

        widget.setupUi(getWidget());

        // setup memory
        this->mem_wrapper = std::make_shared<skills::gui::SkillManagerWrapper>();

        viewer = std::make_unique<armarx::skills::gui::SkillMemoryGUI>(
            widget.treeWidgetSkillExecutions,
            widget.groupBoxSkillExecutions->layout(),

            widget.groupBoxSkills,
            widget.splitter,

            widget.groupBoxSkillDetails,
            widget.splitter,

            widget.updateWidgetLayout,

            widget.skillDescription,
            widget.skillDescription->parentWidget()->layout(),

            this->mem_wrapper);


        /*
        this->updateWidget = new PeriodicUpdateWidget(2.0, 60);
        widget.updateWidgetLayout->insertWidget(0, updateWidget);

        this->stopAllButton = new QPushButton("Stop all executions");
        widget.stopAllLayout->addWidget(stopAllButton);


        skillDescriptionWidget = new SkillDescriptionWidget();
        widget.skillDescription->parentWidget()->layout()->replaceWidget(widget.skillDescription,
                                                                         skillDescriptionWidget);
        widget.skillDescription = skillDescriptionWidget;



        connect(this->stopAllButton,
                &QPushButton::clicked,
                this,
                &SkillManagerMonitorWidgetController::stopAllExecutions);
        connect(widget.treeWidgetSkillExecutions,
                &QTreeWidget::customContextMenuRequested,
                this,
                &SkillManagerMonitorWidgetController::prepareAndRunMenu);

        connect(updateWidget,
                &armarx::PeriodicUpdateWidget::update,
                this,
                &SkillManagerMonitorWidgetController::refreshSkillsAndExecutions);

        connect(widget.pushButtonCopy,
                &QPushButton::clicked,
                this,
                &SkillManagerMonitorWidgetController::copyCurrentConfig);
        connect(widget.pushButtonPaste,
                &QPushButton::clicked,
                this,
                &SkillManagerMonitorWidgetController::pasteCurrentConfig);

        connect(widget.pushButtonExecuteSkill,
                &QPushButton::clicked,
                this,
                &SkillManagerMonitorWidgetController::executeSelectedSkill);

        connect(widget.treeWidgetSkills,
                &QTreeWidget::currentItemChanged,
                this,
                &SkillManagerMonitorWidgetController::skillSelectionChanged);

        connect(widget.treeWidgetSkillExecutions,
                &QTreeWidget::currentItemChanged,
                this,
                &SkillManagerMonitorWidgetController::skillExecutionSelectionChanged);

        connect(widget.pushButtonSearch,
                &QPushButton::clicked,
                this,
                &SkillManagerMonitorWidgetController::searchSkills);
        // alternatively run search when pressing enter
        connect(widget.lineEditSearch,
                &QLineEdit::returnPressed,
                this,
                &SkillManagerMonitorWidgetController::searchSkills);
*/
    }

    void
    SkillManagerMonitorWidgetController::connectGui()
    {
        /*
        connect(this,
                &SkillManagerMonitorWidgetController::disconnect,
                this->mem_wrapper,
                &skills::gui::SkillManagerWrapper::disconnect);
        connect(this,
                &SkillManagerMonitorWidgetController::disconnect,
                this->viewer,
                &skills::gui::SkillMemoryGUI::disconnect);
        */
    }

    SkillManagerMonitorWidgetController::~SkillManagerMonitorWidgetController()
    {
    }

    void
    SkillManagerMonitorWidgetController::onInitComponent()
    {
        usingProxy(observerName);
    }

    void
    SkillManagerMonitorWidgetController::onConnectComponent()
    {
        /*
        widget.groupBoxSkills->setTitle(QString::fromStdString(observerName));
        widget.treeWidgetSkillDetails->setEditTriggers(
            QAbstractItemView::EditTrigger::NoEditTriggers);
        widget.treeWidgetSkillDetails->setColumnHidden(3, true);
*/


        //this->updateWidget->startTimerIfEnabled();

        getProxy(this->memory, observerName, 1000);
        //this->mem_wrapper->connect(memory);
        connected = true;
    }

    void
    SkillManagerMonitorWidgetController::onDisconnectComponent()
    {
        connected = false;
        memory = nullptr;
        this->mem_wrapper->disconnect();

        // reset all
        /*
        skills.clear();
        widget.treeWidgetSkills->clear();
        widget.treeWidgetSkillDetails->clear();
        skillsArgumentsTreeWidgetItem = nullptr;
        selectedSkill.skillId.providerId->providerName = "";
        selectedSkill.skillId.skillName = "";
        //this->updateWidget->stopTimer();
*/
    }

    /*
    void
    SkillManagerMonitorWidgetController::searchSkills()
    {
        this->currentSkillSearch = widget.lineEditSearch->text();
    }

    void
    SkillManagerMonitorWidgetController::refreshSkillsAndExecutions()
    {
        refreshSkills();
        refreshExecutions();
    }

    void
    SkillManagerMonitorWidgetController::matchSkillUpdateToSearch(
        std::map<skills::manager::dto::SkillID, skills::manager::dto::SkillDescription>& update)
    {

        if (this->currentSkillSearch.isEmpty())
        {
            return;
        }

        // TODO: whitespace to wildcard

        std::string search = this->currentSkillSearch.toLower().toStdString();
        std::transform(search.begin(), search.end(), search.begin(),
                       [](unsigned char c){ return std::tolower(c); });

        for (auto it = update.begin(); it != update.end();)
        {
            if (simox::alg::to_lower(skills::SkillID::FromIce(it->first).skillName)
                    .find(this->currentSkillSearch.toLower().toStdString()))
            {
                it = update.erase(it);
            }
            else
            {
                ++it;
            }
        }

    }

    void
    SkillManagerMonitorWidgetController::refreshSkills()
    {
        if (!memory)
        {
            // check if null
            return;
        }


        // remove non-existing ones
        try
        {
            std::scoped_lock l(updateMutex);

            auto managerSkills = memory->getSkillDescriptions();
            this->matchSkillUpdateToSearch(managerSkills);

            // completely recreate internal skills map
            skills.clear();
            for (const auto& [sid, desc] : managerSkills)
            {
                auto description = skills::SkillDescription::FromIce(desc);
                auto skillId = skills::SkillID::FromIce(sid);
                auto providerId = skillId.providerId.value_or(
                    skills::ProviderID{.providerName = "UNKNOWN PROVIDER NAME"});

                ARMARX_CHECK(skillId.isFullySpecified());

                auto& providedSkillsMap = skills[providerId]; // create new if not existent
                providedSkillsMap.insert({skillId, description});
            }

            // update tree view. Remove non-existing elements
            int i = 0;
            while (i < widget.treeWidgetSkills->topLevelItemCount())
            {
                auto* providerItem = widget.treeWidgetSkills->topLevelItem(i);
                auto providerName = providerItem->text(0).toStdString();
                skills::ProviderID providerId{.providerName = providerName};

                if (skills.find(providerId) == skills.end())
                {
                    providerItem = nullptr; // reset
                    auto remove = widget.treeWidgetSkills->takeTopLevelItem(i);
                    delete remove;
                    continue;
                }

                ++i;
                auto& providedSkills = skills.at(providerId);

                int j = 0;
                while (j < providerItem->childCount())
                {
                    auto* skillItem = providerItem->child(j);
                    auto skillName = skillItem->text(0).toStdString();

                    skills::SkillID skillId{.providerId =
                                                skills::ProviderID{.providerName = providerName},
                                            .skillName = skillName};

                    if (providedSkills.find(skillId) == providedSkills.end())
                    {
                        skillItem = nullptr;
                        auto remove = providerItem->takeChild(j);
                        delete remove;
                        continue;
                    }

                    ++j;
                }
            }

            // update tree view. Add new elements
            for (const auto& [providerId, providedSkills] : skills)
            {
                QTreeWidgetItem* providerItem = nullptr;
                for (int i = 0; i < widget.treeWidgetSkills->topLevelItemCount(); ++i)
                {
                    auto el = widget.treeWidgetSkills->topLevelItem(i);
                    auto providerName = el->text(0).toStdString();
                    skills::ProviderID elProviderId{.providerName = providerName};

                    if (providerId == elProviderId)
                    {
                        providerItem = el;
                        break;
                    }
                }

                if (!providerItem)
                {
                    providerItem = new QTreeWidgetItem(widget.treeWidgetSkills);
                    providerItem->setText(0, QString::fromStdString(providerId.providerName));
                }

                for (const auto& [skillId, skill] : providedSkills)
                {
                    QTreeWidgetItem* skillItem = nullptr;
                    for (int i = 0; i < providerItem->childCount(); ++i)
                    {
                        auto el = providerItem->child(i);
                        auto skillName = el->text(0).toStdString();
                        skills::SkillID elSkillId{providerId, skillName};

                        if (skillId == elSkillId)
                        {
                            skillItem = el;
                            break;
                        }
                    }

                    if (!skillItem)
                    {
                        skillItem = new SkillInfoTreeWidgetItem(skill, providerItem);
                        skillItem->setText(0, QString::fromStdString(skillId.skillName));
                    }
                }
            }
        }
        catch (...)
        {
            // perhaps the manager died during the method?
        }
    }

    void
    SkillManagerMonitorWidgetController::refreshExecutions()
    {
        static std::map<skills::SkillStatus, std::string> ExecutionStatus2String = {
            // Main states
            {skills::SkillStatus::Constructing, "Constructing"},
            {skills::SkillStatus::Initializing, "Initializing"},
            {skills::SkillStatus::Preparing, "Preparing"},
            {skills::SkillStatus::Running, "Running"},

            // Terminating
            {skills::SkillStatus::Aborted, "Aborted"},
            {skills::SkillStatus::Failed, "Failed"},
            {skills::SkillStatus::Succeeded, "Succeeded"}};

        if (!memory)
        {
            // check if null
            return;
        }

        try
        {
            std::scoped_lock l(updateMutex);

            auto currentManagerStatuses = memory->getSkillExecutionStatuses();

            for (const auto& [k, v] : currentManagerStatuses)
            {
                auto executionId = skills::SkillExecutionID::FromIce(k);
                auto statusUpdate = skills::SkillStatusUpdate::FromIce(v);

                // update maps
                skillStatusUpdates[executionId] = statusUpdate;
                //skillExecutionParams.insert_or_assign(executionId,
                //                                      statusUpdate.usedParameterization);

                SkillExecutionInfoTreeWidgetItem* found = nullptr;
                for (int i = 0; i < widget.treeWidgetSkillExecutions->topLevelItemCount(); ++i)
                {
                    auto c = static_cast<SkillExecutionInfoTreeWidgetItem*>(
                        widget.treeWidgetSkillExecutions->topLevelItem(i));

                    found =
                        SkillExecutionInfoTreeWidgetItem::SearchRecursiveForMatch(c, executionId);

                    if (found)
                    {
                        for (std::pair<skills::SkillStatus, std::string> i : ExecutionStatus2String)
                        {
                            if (i.first == statusUpdate.status)
                            {
                                found->setText(3, QString::fromStdString(i.second));
                            }
                        }

                        break;
                    }
                }

                if (!found)
                {
                    // TODO: Sort to executor!
                    auto item = new SkillExecutionInfoTreeWidgetItem(executionId);

                    item->setText(0,
                                  QString::fromStdString(
                                      executionId.executionStartedTime.toDateTimeString()));
                    item->setText(1, QString::fromStdString(executionId.executorName));
                    item->setText(2, QString::fromStdString(executionId.skillId.toString()));
                    for (std::pair<skills::SkillStatus, std::string> i : ExecutionStatus2String)
                    {
                        if (i.first == statusUpdate.status)
                        {
                            item->setText(3, QString::fromStdString(i.second));
                        }
                    }

                    widget.treeWidgetSkillExecutions->insertTopLevelItem(0, item);
                }
            }
        }
        catch (...)
        {
            // perhaps the manager died during the method?
        }
    }

    void
    SkillManagerMonitorWidgetController::executeSelectedSkill()
    {
        auto data = getConfigAsAron();
        executeSkillWithParams(selectedSkill.skillId, data);
    }

    void
    SkillManagerMonitorWidgetController::rerunSkillWithSimilarParams()
    {
        // TODO: disabled until skillparameterization is replaced
        //skills::SkillParameterization selectedExecutionParams =
        //skillExecutionParams.at(selectedSkill.skillExecutionId);
        //executeSkillWithParams(selectedSkill.skillExecutionId.skillId,
        //                       selectedExecutionParams.parameterization);

        executeSelectedSkill();
    }


    void
    SkillManagerMonitorWidgetController::executeSkillWithParams(skills::SkillID skillId,
                                                                aron::data::DictPtr params)
    {
        std::scoped_lock l(updateMutex);

        auto providerId = skillId.providerId;
        if (!providerId.has_value())
        {
            return;
        }
        const auto& skillDescriptions = skills.at(providerId.value());
        if (skillDescriptions.find(skillId) == skillDescriptions.end())
        {
            return;
        }

        char hostname[HOST_NAME_MAX];
        gethostname(hostname, HOST_NAME_MAX);

        skills::SkillExecutionRequest req{
            .skillId = selectedSkill.skillId,
            .executorName = "Skills.Manager GUI (hostname: " + std::string(hostname) + ")",
            .parameters = params};

        ARMARX_CHECK(selectedSkill.skillId.isFullySpecified()); // sanity check
        ARMARX_IMPORTANT << "Executing skill from GUI: " << selectedSkill.skillId << ".";
        // Note that we execute the skill in a seperate thread so that the GUI thread does not freeze.
        memory->begin_executeSkill(req.toManagerIce());
    }

    void
    SkillManagerMonitorWidgetController::stopAllExecutions()
    {
        QTreeWidget const* tree = widget.treeWidgetSkillExecutions;

        int const max_retries = 3;
        int left_retries = max_retries;
        bool retry = false;

        do
        {
            retry = false;

            for (ssize_t i = 0; i < tree->topLevelItemCount(); ++i)
            {
                SkillExecutionInfoTreeWidgetItem const* item =
                    dynamic_cast<SkillExecutionInfoTreeWidgetItem*>(
                        widget.treeWidgetSkillExecutions->topLevelItem(i));

                if (not item)
                {
                    // At this point, dynamic-casting failed and we don't know anything about the
                    // type.
                    ARMARX_ERROR << "Cannot stop unknown skill.";
                    retry = true;
                    continue;
                }

                try
                {
                    ARMARX_INFO << "Aborting skill '" << item->executionId.skillId.skillName
                                << "'...";
                    memory->abortSkillAsync(item->executionId.toManagerIce());
                }
                catch (Ice::Exception const& e)
                {
                    retry = true;
                    ARMARX_ERROR << "Unhandeled Ice exception while aborting skill '"
                                 << item->executionId.skillId.skillName << "'.";
                }
                catch (...)
                {
                    retry = true;
                    ARMARX_ERROR << "Unhandled error while aborting skill '"
                                 << item->executionId.skillId.skillName << "'.";
                }
            }

            if (retry)
            {
                left_retries -= 1;

                if (left_retries > 0)
                {
                    ARMARX_WARNING << "There where errors aborting skills.  Retrying...";
                }
                else
                {
                    ARMARX_ERROR << "Couldn't abort all skills after " << max_retries
                                 << " tries.  Giving up.";
                    retry = false;
                }
            }
        } while (retry);
    }

    void
    SkillManagerMonitorWidgetController::stopSkill()
    {
        std::scoped_lock l(updateMutex);

        if (selectedSkill.skillExecutionId.skillId.fullySpecified())
        {
            ARMARX_INFO << "The user requested to stop a skill, which was not fully specified!";
            return;
        }

        const auto& skillDescriptions = skills.at(*selectedSkill.skillId.providerId);
        if (!skillDescriptions.count(selectedSkill.skillId.skillName))
        {
            return;
        }

        ARMARX_INFO << "Stopping skill from GUI: " << selectedSkill.skillExecutionId.toString();

        memory->abortSkillAsync(selectedSkill.skillExecutionId.toManagerIce());
    }

    void
    SkillManagerMonitorWidgetController::skillExecutionSelectionChanged(QTreeWidgetItem* current,
                                                                        QTreeWidgetItem*)
    {
        std::scoped_lock l(updateMutex);
        widget.groupBoxSkillExecutions->setEnabled(false);

        if (!current)
        {
            // gui has died?
            return;
        }

        auto c = static_cast<SkillExecutionInfoTreeWidgetItem*>(current);
        selectedSkill.skillExecutionId = c->executionId;
        widget.groupBoxSkillExecutions->setEnabled(true);
    }

    void
    SkillManagerMonitorWidgetController::skillSelectionChanged(QTreeWidgetItem* current,
                                                               QTreeWidgetItem*)
    {
        std::scoped_lock l(updateMutex);
        widget.groupBoxSkillDetails->setEnabled(false);

        if (!current)
        {
            // gui has died?
            return;
        }

        if (!current->parent())
        {
            // no parent available. Perhaps provider clicked? Reset selected skill.
            selectedSkill.skillId = SelectedSkill::UNK_SKILL_ID;
            return;
        }

        auto c = static_cast<SkillInfoTreeWidgetItem*>(current);
        auto skillDescription = c->skillDescription;


        if (selectedSkill.skillId == skillDescription.skillId)
        {
            // no change
            return;
        }

        selectedSkill.skillId = skillDescription.skillId;

        // setup groupBox
        widget.groupBoxSkillDetails->setTitle(
            QString::fromStdString(selectedSkill.skillId.toString()));
        widget.groupBoxSkillDetails->setEnabled(true);

        // setup table view
        widget.treeWidgetSkillDetails->clear();
        aronTreeWidgetController = nullptr;
        skillsArgumentsTreeWidgetItem = nullptr;

        // We assert that the skill exists
        ARMARX_CHECK(skills.count(*selectedSkill.skillId.providerId) > 0);
        ARMARX_CHECK(skills.at(*selectedSkill.skillId.providerId).count(selectedSkill.skillId) > 0);
        auto skillDesc = skills.at(*selectedSkill.skillId.providerId).at(selectedSkill.skillId);

        {
            auto it = new QTreeWidgetItem(widget.treeWidgetSkillDetails,
                                          {QString::fromStdString("Name"),
                                           QString::fromStdString(skillDesc.skillId.skillName)});
            widget.treeWidgetSkillDetails->addTopLevelItem(it);
        }

        {
            auto it = new QTreeWidgetItem(widget.treeWidgetSkillDetails,
                                          {QString::fromStdString("Description"),
                                           QString::fromStdString(skillDesc.description)});
            widget.treeWidgetSkillDetails->addTopLevelItem(it);
        }

        {
            auto it = new QTreeWidgetItem(
                widget.treeWidgetSkillDetails,
                {QString::fromStdString("Timeout"),
                 QString::fromStdString(std::to_string(skillDesc.timeout.toMilliSeconds())) +
                     " ms"});
            widget.treeWidgetSkillDetails->addTopLevelItem(it);
        }

        // select root profile
        widget.comboBoxProfiles->setCurrentIndex(0);

        // remove any profile
        while (widget.comboBoxProfiles->count() > 1)
        {
            widget.comboBoxProfiles->removeItem(1);
        }

        // add new profiles for this skill
        // TODO: Where stored?

        skillsArgumentsTreeWidgetItem = new QTreeWidgetItem(widget.treeWidgetSkillDetails,
                                                            {QString::fromStdString("Arguments")});
        auto aron_args = skillDesc.parametersType;
        auto default_args_of_profile = skillDesc.rootProfileDefaults;

        aronTreeWidgetController =
            std::make_shared<AronTreeWidgetController>(widget.treeWidgetSkillDetails,
                                                       skillsArgumentsTreeWidgetItem,
                                                       aron_args,
                                                       default_args_of_profile);

        // automatically expand args
        skillsArgumentsTreeWidgetItem->setExpanded(true);
    }

    aron::data::DictPtr
    SkillManagerMonitorWidgetController::getConfigAsAron() const
    {
        // create argument aron (if there is an accepted type set)
        if (aronTreeWidgetController)
        {
            return aronTreeWidgetController->convertToAron();
        }
        return nullptr;
    }

    void
    SkillManagerMonitorWidgetController::copyCurrentConfig()
    {
        auto data = getConfigAsAron();
        if (!data)
        {
            return;
        }

        auto json = aron::data::converter::AronNlohmannJSONConverter::ConvertToNlohmannJSON(data);
        QClipboard* clipboard = QApplication::clipboard();
        clipboard->setText(QString::fromStdString(json.dump(2)));
    }

    void
    SkillManagerMonitorWidgetController::pasteCurrentConfig()
    {
        QClipboard* clipboard = QApplication::clipboard();
        std::string s = clipboard->text().toStdString();
        nlohmann::json json = nlohmann::json::parse(s);
        auto data =
            aron::data::converter::AronNlohmannJSONConverter::ConvertFromNlohmannJSONObject(json);

        if (!aronTreeWidgetController)
        {
            return;
        }

        aronTreeWidgetController->setFromAron(data);
    }

    void
    SkillManagerMonitorWidgetController::resetCurrentConfig()
    {
        // TODO
    }
*/

} // namespace armarx
