#include "AronTreeWidgetController.h"

#include "RobotAPI/libraries/skills_gui/aron_tree_widget/visitors/AronTreeWidgetModalCreator.h"

#include "visitors/AronTreeWidgetContextMenu.h"
#include "visitors/AronTreeWidgetConverter.h"
#include "visitors/AronTreeWidgetSetter.h"

namespace armarx::skills::gui
{
    AronTreeWidgetController::AronTreeWidgetController(QTreeWidget* tree,
                                                       QTreeWidgetItem* parent,
                                                       const aron::type::ObjectPtr& type,
                                                       const aron::data::DictPtr& data) :
        parent(parent), tree(tree), type(type)
    {
        connect(tree,
                SIGNAL(customContextMenuRequested(const QPoint&)),
                this,
                SLOT(ShowContextMenu(const QPoint&)));
        connect(tree,
                &QTreeWidget::itemDoubleClicked,
                this,
                &AronTreeWidgetController::onTreeWidgetItemDoubleClicked);

        if (type) // if there is a type set, we create a tree widget from the type
        {
            AronTreeWidgetCreatorVisitor v(parent);
            v.setTopLevelWidget(tree);
            aron::type::visit(v, type);

            if (data) // check if there is a default argument set. Prefill the GUI with it
            {
                setFromAron(data);
            }
        }
        // there is no type but a default configuration. Prefill the GUI with the default arguments
        else if (data)
        {
            // TODO: There is no visitor for that (yet)...
            // create type from data, ...
        }
        else
        {
            new QTreeWidgetItem(parent, {QString::fromStdString("No args")});
        }

        // connect change handling after args init
        connect(tree,
                &QTreeWidget::itemChanged,
                this,
                &AronTreeWidgetController::onTreeWidgetItemChanged);
    }

    aron::data::DictPtr
    AronTreeWidgetController::convertToAron() const
    {
        if (parent && type)
        {
            AronTreeWidgetConverterVisitor v(parent, 0);
            aron::type::visit(v, type);

            auto aron_args = aron::data::Dict::DynamicCastAndCheck(v.createdAron);
            return aron_args;
        }
        return nullptr;
    }

    void
    AronTreeWidgetController::setFromAron(const aron::data::DictPtr& data)
    {
        if (parent)
        {
            AronTreeWidgetSetterVisitor v(parent, 0);
            aron::data::visit(v, data);
        }
    }

    void
    AronTreeWidgetController::ShowContextMenu(const QPoint& pos)
    {
        tree->blockSignals(true);

        auto idx = tree->indexAt(pos);
        AronTreeWidgetItem* clickedItem = AronTreeWidgetItem::DynamicCast(tree->itemAt(pos));
        if (clickedItem)
        {
            AronTreeWidgetContextMenuVisitor visitor(clickedItem, pos, tree, idx.row());

            aron::type::visit(visitor, clickedItem->aronType);
            visitor.showMenuAndExecute();
        }

        tree->blockSignals(false);
    }

    void
    AronTreeWidgetController::onTreeWidgetItemDoubleClicked(QTreeWidgetItem* item, int column)
    {
        if (!item)
        {
            return;
        }
        tree->blockSignals(true);

        auto* aronItem = AronTreeWidgetItem::DynamicCast(item);
        if (column == 1 && aronItem)
        {


            std::string name =
                aronItem->text(aron_tree_widget::constantes::TREE_WIDGET_ITEM_NAME).toStdString();
            // depending on aron type, create extra gui element.
            armarx::skills::gui::AronTreeWidgetModalCreatorVisitor v(name, aronItem, tree);
            aron::type::visit(v, aronItem->aronType);
            auto modal = v.createdModal;

            // if no modal is created, we instead use the edit field directly
            if (modal)
            {
                modal->exec();
            }
            else if (aronItem->col1Editable)
            {
                item->treeWidget()->editItem(item, column);
            }
        }
        else if (column == 0 && aronItem && aronItem->col0Editable)
        {
            item->treeWidget()->editItem(item, column);
        }

        tree->blockSignals(false);
    }

    void
    AronTreeWidgetController::onTreeWidgetItemChanged(QTreeWidgetItem* item, int column)
    {
        tree->blockSignals(true);

        auto* aronElem = AronTreeWidgetItem::DynamicCast(item);
        if (aronElem)
        {
            aronElem->onUserChange(column);
        }
        // start conversion for entire tree -- this also sets the highlighting
        if (parent->childCount() == 1)
        {
            auto* aronTreeRoot = AronTreeWidgetItem::DynamicCast(parent->child(0));
            aronTreeRoot->resetError();
            if (aronTreeRoot)
            {
                AronTreeWidgetConverterVisitor v(parent, 0);
                aron::type::visit(v, type);
                aronTreeRoot->setValueErrorState(v.hasDirectError(), v.onlyChildFailedConversion());
            }
        }
        // else perhaps the GUI was stopped or died.

        tree->blockSignals(false);
    }

} // namespace armarx::skills::gui
