#include "SkillExecutionTreeWidget.h"

#include <mutex>

#include <QMenu>
#include <QTreeWidgetItem>

#include "SkillExecutionTreeWidgetItem.h"

namespace armarx::skills::gui
{

    void
    SkillExecutionTreeWidget::runContextMenu(const QPoint& pos)
    {
        QMenu* menu = new QMenu();

        // Stop skill
        QAction* stopSkillAction = new QAction("Stop execution", this);
        skills::SkillStatus currentStatus =
            memory->getLatestUpdate().statuses.at(selectedExecution.skillExecutionId).status;
        stopSkillAction->setDisabled(currentStatus == skills::SkillStatus::Aborted ||
                                     currentStatus == skills::SkillStatus::Failed ||
                                     currentStatus == skills::SkillStatus::Succeeded);

        QAction* rerunSkillAction = new QAction("Re-run with similar params", this);
        menu->addAction(stopSkillAction);
        menu->addAction(rerunSkillAction);
        connect(stopSkillAction,
                &QAction::triggered,
                this,
                &SkillExecutionTreeWidget::stopSelectedExecution);
        connect(rerunSkillAction,
                &QAction::triggered,
                this,
                &SkillExecutionTreeWidget::rerunSkillWithSimilarParams);

        // Temporarily disable rerun-skill-Action
        rerunSkillAction->setDisabled(true);

        // open menu
        menu->popup(this->viewport()->mapToGlobal(pos));
    }

    void
    SkillExecutionTreeWidget::stopSelectedExecution()
    {
        if (!selectionValid())
            return;
        memory->stopExecution(this->selectedExecution.skillExecutionId);
    }

    void
    SkillExecutionTreeWidget::rerunSkillWithSimilarParams()
    {
        if (!selectionValid())
            return;
        // we don't want to hold state in the gui, so we need to get the parameters from memory:
        skills::SkillExecutionID currentExecutionId = this->selectedExecution.skillExecutionId;
        auto update = memory->getLatestUpdate();
        if (!update.statuses.contains(currentExecutionId))
        {
            // we didn't find an entry for the execution id
            ARMARX_IMPORTANT << "The selected execution was not found in memory. The GUI is unable "
                                "to determine the parametrization for this execution.";
            return;
        }
        auto params = update.statuses[currentExecutionId].parameters;

        // give all information to manager
        this->memory->startExecutionWithParams(currentExecutionId.skillId, params);
    }

    void
    SkillExecutionTreeWidget::setupUi()
    {
        this->setColumnCount(4);
    }

    inline bool
    SkillExecutionTreeWidget::selectionValid()
    {
        return selectedExecution.skillExecutionId.skillId.skillName != skills::SkillID::UNKNOWN;
    }

    void
    SkillExecutionTreeWidget::stopAllExecutions()
    {
        for (ssize_t i = 0; i < this->topLevelItemCount(); ++i)
        {
            auto item = dynamic_cast<SkillExecutionTreeWidgetItem*>(this->topLevelItem(i));

            if (!item)
            {
                continue;
            }

            memory->stopExecution(item->getExecutionId());
        }
    }

    void
    SkillExecutionTreeWidget::executionSelectionChanged(QTreeWidgetItem* current,
                                                        QTreeWidgetItem* previous)
    {
        // update internal state
        SkillExecutionTreeWidgetItem* selected =
            dynamic_cast<SkillExecutionTreeWidgetItem*>(current);
        if (selected)
        {
            this->selectedExecution.skillExecutionId = selected->getExecutionId();
        }
    }

    void
    SkillExecutionTreeWidget::refresh()
    {
        auto currentManagerStatuses = memory->getLatestUpdate().statuses;

        for (const auto& [k, v] : currentManagerStatuses)
        {
            skills::SkillExecutionID executionId = k;
            skills::SkillStatusUpdate statusUpdate = v;

            SkillExecutionTreeWidgetItem* found = nullptr;
            for (int i = 0; i < this->topLevelItemCount(); ++i)
            {
                auto c = dynamic_cast<SkillExecutionTreeWidgetItem*>(topLevelItem(i));
                if (!c)
                {
                    // the item is probably not the correct type, skip...
                    continue;
                }

                found = SkillExecutionTreeWidgetItem::SearchRecursiveForMatch(c, executionId);

                if (found)
                {
                    for (std::pair<skills::SkillStatus, std::string> i : EXECUTION_STATUS_TO_STRING)
                    {
                        if (i.first == statusUpdate.status)
                        {
                            found->setText(3, QString::fromStdString(i.second));
                        }
                    }

                    break;
                }
            }

            if (!found)
            {
                // TODO: Sort to executor!
                auto item = new SkillExecutionTreeWidgetItem(executionId, memory);

                item->setText(
                    0, QString::fromStdString(executionId.executionStartedTime.toDateTimeString()));
                item->setText(1, QString::fromStdString(executionId.executorName));
                item->setText(2, QString::fromStdString(executionId.skillId.toString()));
                for (std::pair<skills::SkillStatus, std::string> i : EXECUTION_STATUS_TO_STRING)
                {
                    if (i.first == statusUpdate.status)
                    {
                        item->setText(3, QString::fromStdString(i.second));
                    }
                }

                this->insertTopLevelItem(0, item);
            }
        }
    }
} // namespace armarx::skills::gui
