#include "SkillManagerWrapper.h"

#include <mutex>

#include "RobotAPI/libraries/skills/core/SkillExecutionRequest.h"

namespace armarx::skills::gui
{
    void
    SkillManagerWrapper::fetchUpdates()
    {
        fetchSkills();
        fetchExecutions();
    }

    void
    SkillManagerWrapper::fetchExecutions()
    {
        ARMARX_ERROR << "Function not implemented!";
    }

    void
    SkillManagerWrapper::filterSkillUpdate(
        std::map<skills::manager::dto::SkillID, skills::manager::dto::SkillDescription>& update)
    {
        // empty search => do not filter
        if (this->currentSkillSearch.empty())
        {
            return;
        }

        // TODO: whitespace to wildcard

        std::string key = simox::alg::to_lower(this->currentSkillSearch);

        for (auto it = update.begin(); it != update.end();)
        {
            if (simox::alg::to_lower(skills::SkillID::FromIce(it->first).skillName).find(key))
            {
                it = update.erase(it);
            }
            else
            {
                ++it;
            }
        }
    }

    void
    SkillManagerWrapper::fetchSkills()
    {
        if (!memory)
        {
            return;
        }

        try
        {
            std::scoped_lock l(mutex_memory);

            auto managerSkills = memory->getSkillDescriptions();

            this->filterSkillUpdate(managerSkills);

            // completely recreate internal skills map
            snapshot.skills.clear();
            for (const auto& [sid, desc] : managerSkills)
            {
                auto description = skills::SkillDescription::FromIce(desc);
                auto skillId = skills::SkillID::FromIce(sid);
                auto providerId = skillId.providerId.value_or(
                    skills::ProviderID{.providerName = "UNKNOWN PROVIDER NAME"});

                ARMARX_CHECK(skillId.isFullySpecified());

                auto& providedSkillsMap = snapshot.skills[providerId]; // create new if not existent
                providedSkillsMap.insert({skillId, description});
            }
        }
        catch (Ice::Exception const& e)
        {
            ARMARX_WARNING
                << "Unhandled Ice exception encountered while updating skills. Exception was: "
                << e;
            return;
        }
        catch (...)
        {
            ARMARX_WARNING << "Unknown exception encountered while updating skills.";
            return;
        }
    }

    void
    SkillManagerWrapper::connect(skills::manager::dti::SkillManagerInterfacePrx& updatedMemory)
    {
        std::scoped_lock l(mutex_memory);
        this->memory = updatedMemory;
    }

    void
    SkillManagerWrapper::disconnect()
    {
        // clear memory
        {
            std::scoped_lock l(mutex_memory);
            this->memory = nullptr;
        }

        // clear stored data
        {
            std::scoped_lock d(mutex_snapshot);
            this->snapshot.skills.clear();
            this->snapshot.statuses.clear();
        }
    }

    void
    SkillManagerWrapper::acceptSearchRequest(std::string search)
    {
        this->currentSkillSearch = search;
    }

    const SkillManagerWrapper::Snapshot
    SkillManagerWrapper::getLatestUpdate()
    {
        // We *want* to make a const copy, since a reference to the snapshot would not necessarily be thread safe.
        // If this is too slow, this class might implement wrapper functions for the maps.
        std::scoped_lock l(mutex_snapshot);
        return snapshot;
    }

    void
    SkillManagerWrapper::stopExecution(skills::SkillExecutionID const& executionId,
                                       unsigned int max_retries)
    {
        // memory???
        if (!memory)
        {
            return;
        }

        unsigned int retries = max_retries;
        bool retry = false;

        do
        {
            try
            {
                ARMARX_INFO << "Aborting skill '" << executionId.skillId.skillName << "'...";
                this->memory->abortSkillAsync(executionId.toManagerIce());
            }
            catch (Ice::Exception const& e)
            {
                retry = true;
                ARMARX_ERROR << "Unhandeled Ice exception while aborting skill '"
                             << executionId.skillId.skillName << "'.";
            }
            catch (...)
            {
                retry = true;
                ARMARX_ERROR << "Unhandled error while aborting skill '"
                             << executionId.skillId.skillName << "'.";
            }

            if (retry)
            {
                retries -= 1;

                if (retries > 0)
                {
                    ARMARX_WARNING << "There where errors aborting skills.  Retrying...";
                }
                else
                {
                    ARMARX_ERROR << "Couldn't abort all skills after " << max_retries
                                 << " tries.  Giving up.";
                    retry = false;
                }
            }
        } while (retry);
    }

    void
    SkillManagerWrapper::startExecutionWithParams(skills::SkillID const& skillId,
                                                  aron::data::DictPtr params)
    {
        std::scoped_lock l(mutex_memory);

        // Memory???
        if (!memory)
        {
            return;
        }

        auto providerId = skillId.providerId;
        if (!providerId.has_value())
        {
            ARMARX_WARNING << "The skill: '" << skillId.skillName
                           << "' has been requested to be executed, but no provider was "
                              "given. Aborting...";
            return;
        }

        const auto& skillDescriptions = snapshot.skills.at(providerId.value());
        if (skillDescriptions.find(skillId) == skillDescriptions.end())
        {
            ARMARX_INFO << "The Skill: '" << skillId.skillName
                        << "' has been requested to be executed, but no skill description was "
                           "found. Aborting...";
            return;
        }

        char hostname[HOST_NAME_MAX];
        gethostname(hostname, HOST_NAME_MAX);

        skills::SkillExecutionRequest req{
            .skillId = skillId,
            .executorName = "Skills.Manager GUI (hostname: " + std::string(hostname) + ")",
            .parameters = params};

        ARMARX_CHECK(skillId.isFullySpecified()); // sanity check
        ARMARX_IMPORTANT << "Executing skill from GUI: " << skillId << ".";

        try
        {
            memory->executeSkill(req.toManagerIce());
        }
        catch (Ice::Exception const& e)
        {
            ARMARX_ERROR << "Unhandeled Ice exception while executing skill '" << skillId.skillName
                         << "'. Aborting...";
        }
        catch (...)
        {
            ARMARX_ERROR << "Unhandled error while executing skill '" << skillId.skillName
                         << "'. Aborting...";
        }
    }


} // namespace armarx::skills::gui
