#pragma once

#include <mutex>

#include <QObject>

#include "RobotAPI/libraries/skills/core/ProviderID.h"
#include "RobotAPI/libraries/skills/core/SkillDescription.h"
#include "RobotAPI/libraries/skills/core/SkillExecutionID.h"
#include "RobotAPI/libraries/skills/core/SkillStatusUpdate.h"
#include <RobotAPI/interface/skills/SkillManagerInterface.h>

namespace armarx::skills::gui
{
    class SkillManagerWrapper : public QObject
    {

    public:
        SkillManagerWrapper(skills::manager::dti::SkillManagerInterfacePrx& _memory,
                            QObject* parent = nullptr) :
            QObject(parent), memory(_memory)
        {
        }

        // We can instantiate the proxy without memory, if it's not connected yet.
        SkillManagerWrapper(QObject* parent = nullptr) : QObject(parent)
        {
        }

        /**
         * @brief Updates the memory pointer. This should be called whenever the GUI connects/reconnects.
         */
        void connect(skills::manager::dti::SkillManagerInterfacePrx& updatedMemory);

        /**
         * @brief The most current snapshot of the memory. Updated with each memory fetch.
         */
        struct Snapshot
        {
            std::map<skills::SkillExecutionID, skills::SkillStatusUpdate> statuses;

            std::map<skills::ProviderID, std::map<skills::SkillID, skills::SkillDescription>>
                skills;

            // Default constructor
            Snapshot() : statuses({}), skills({})
            {
            }
        };

        // Basic functions to work with. These will produce Ice calls!
        /**
         * @brief Attempts to start an execution with given parameters. If an Ice exception is encountered, a message will be logged and the method will abort.
         * @param skillId The skill to execute.
         * @param params The parameters to execute with.
         */
        void startExecutionWithParams(skills::SkillID const& skillId, aron::data::DictPtr params);

        /**
         * @brief Attempts to stop an execution. If this doesn't succeed, it will retry.
         * @param executionId The ExecutionID to abort.
         * @param max_retries The amount of times to retry, if aborting fails.
         */
        void stopExecution(skills::SkillExecutionID const& executionId,
                           unsigned int max_retries = 0);

        /**
         * @brief Returns a snapshot, which contains the most current update from memory. Calling this function will *not* use an Ice call! Use this instead of holding the state in a widget.
         * @return The most current snapshot
         */
        const Snapshot getLatestUpdate();


    public slots:
        /**
         * @brief Fetches an update from memory. This produces multiple Ice calls!
         */
        void fetchUpdates();

        /**
         * @brief Disconnects the interface from memory.
         */
        void disconnect();

        /**
         * @brief Applies the search word to the update filter.
         * @param search The search word.
         */
        void acceptSearchRequest(std::string search);

    private:
        mutable std::mutex mutex_memory;
        mutable std::mutex mutex_snapshot;

        armarx::skills::manager::dti::SkillManagerInterfacePrx memory;
        Snapshot snapshot;
        std::string currentSkillSearch = "";

        /**
         * @brief Updates the skills map in the snapshot.
         */
        void fetchSkills();

        /**
         * @brief Updates the executions map in the snapshot.
         */
        void fetchExecutions();

        /**
         * @brief Modifies the given map to match the search.
         * @param update The map to modify.
         */
        void filterSkillUpdate(std::map<skills::manager::dto::SkillID,
                                        skills::manager::dto::SkillDescription>& update);
    };
} // namespace armarx::skills::gui
